import logging
import os
import subprocess

from .util import (
    TESTSDIR,
    get_content_checksum,
    needs_sshd,
    random_commit,
    wipe_repo,
)

ostree_receive_abspath = os.path.join(TESTSDIR, 'ostree-receive')
logger = logging.getLogger(__name__)

# Skip all tests here if the required sshd is not available.
pytestmark = needs_sshd


def run_push(source_repo, dest_repo, sshd, ssh_options, env_vars,
             receive_config_path, command='ostree-receive', dest=None,
             options=None, refs=None, **popen_kwargs):
    dest = dest or f'ssh://{sshd.address}:{sshd.port}/{dest_repo.path}'
    options = options or []
    refs = refs or []
    env = os.environ.copy()
    env['OSTREE_RECEIVE_CONF'] = receive_config_path
    if env_vars:
        env.update(env_vars)
    popen_kwargs['env'] = env
    if 'check' not in popen_kwargs:
        popen_kwargs['check'] = True
    cmd = [
        'ostree-push',
        f'--repo={source_repo.path}',
        f'--command={command}',
    ] + ssh_options + options + [dest] + refs
    logger.debug('push command: %s', ' '.join(cmd))
    return subprocess.run(cmd, **popen_kwargs)


def test_no_commits(source_repo, dest_repo, sshd, ssh_options,
                    cli_env_vars, receive_config_path, capfd):
    """Test push with no commits in source repo"""
    args = (
        source_repo, dest_repo, sshd, ssh_options, cli_env_vars,
        receive_config_path
    )

    run_push(*args)
    capfd.readouterr()
    _, receive_refs = dest_repo.list_refs()
    assert receive_refs == {}

    ret = run_push(*args, refs=['foo', 'bar'], check=False)
    _, err = capfd.readouterr()
    assert ret != 0
    assert 'otpush.push.OTPushError: Refs bar foo not found' in err
    _, receive_refs = dest_repo.list_refs()
    assert receive_refs == {}


def test_basic(source_repo, dest_repo, sshd, ssh_options, cli_env_vars,
               receive_config_path, tmp_files_path, capfd):
    """Test push with one commit in source repo"""
    args = (
        source_repo, dest_repo, sshd, ssh_options, cli_env_vars,
        receive_config_path
    )

    rev = random_commit(source_repo, tmp_files_path, 'test')
    source_content = get_content_checksum(source_repo, rev)

    wipe_repo(dest_repo)
    run_push(*args)
    capfd.readouterr()
    _, receive_refs = dest_repo.list_refs()
    assert receive_refs.keys() == {'test', 'ostree-metadata'}
    receive_content = get_content_checksum(dest_repo, receive_refs['test'])
    assert receive_content == source_content

    wipe_repo(dest_repo)
    run_push(*args, refs=['test'])
    capfd.readouterr()
    _, receive_refs = dest_repo.list_refs()
    assert receive_refs.keys() == {'test', 'ostree-metadata'}
    receive_content = get_content_checksum(dest_repo, receive_refs['test'])
    assert receive_content == source_content

    wipe_repo(dest_repo)
    ret = run_push(*args, refs=['test', 'foo'], check=False)
    _, err = capfd.readouterr()
    assert ret != 0
    assert 'otpush.push.OTPushError: Refs foo not found' in err
    _, receive_refs = dest_repo.list_refs()
    assert receive_refs == {}


def test_dry_run(source_repo, dest_repo, sshd, ssh_options, cli_env_vars,
                 receive_config_path, tmp_files_path):
    """Test push dry run"""
    args = (
        source_repo, dest_repo, sshd, ssh_options, cli_env_vars,
        receive_config_path
    )

    random_commit(source_repo, tmp_files_path, 'test')

    wipe_repo(dest_repo)
    run_push(*args, options=['-n'])
    _, receive_refs = dest_repo.list_refs()
    assert receive_refs == {}

    wipe_repo(dest_repo)
    run_push(*args, options=['--dry-run'])
    _, receive_refs = dest_repo.list_refs()
    assert receive_refs == {}

    wipe_repo(dest_repo)
    run_push(*args, options=['-n'], refs=['test'])
    _, receive_refs = dest_repo.list_refs()
    assert receive_refs == {}


def test_scp_dest(source_repo, dest_repo, sshd, ssh_options, cli_env_vars,
                  receive_config_path, tmp_files_path):
    """Test push with scp style destination"""
    args = (
        source_repo, dest_repo, sshd, ssh_options, cli_env_vars,
        receive_config_path
    )
    dest = f'{sshd.address}:{dest_repo.path}'
    options = ['-p', str(sshd.port)]

    random_commit(source_repo, tmp_files_path, 'test')
    run_push(*args, dest=dest, options=options)
    _, receive_refs = dest_repo.list_refs()
    assert receive_refs.keys() == {'test', 'ostree-metadata'}


def test_command_abspath(source_repo, dest_repo, sshd, ssh_options,
                         cli_env_vars, receive_config_path, tmp_files_path):
    """Test push with absolute path to ostree-receive"""
    args = (
        source_repo, dest_repo, sshd, ssh_options, cli_env_vars,
        receive_config_path
    )
    random_commit(source_repo, tmp_files_path, 'test')
    run_push(*args, command=ostree_receive_abspath)
    _, receive_refs = dest_repo.list_refs()
    assert receive_refs.keys() == {'test', 'ostree-metadata'}



import json
import logging
import os
import subprocess

from .util import (
    SRCDIR,
    TESTSDIR,
    needs_sshd,
)

logger = logging.getLogger(__name__)

# Skip all tests here if the required sshd is not available.
pytestmark = needs_sshd


def test_basic(sshd, ssh_options):
    cmd = (
        ['ssh', '-p', str(sshd.port)] +
        ssh_options +
        [sshd.address, 'dumpenv']
    )
    logger.debug('SSH command: %s', ' '.join(cmd))
    logger.debug('Source PATH=%s', os.getenv('PATH'))
    out = subprocess.check_output(cmd)
    data = json.loads(out.decode('utf-8'))

    args = data['args']
    assert args == [os.path.join(TESTSDIR, 'dumpenv')]

    env = data['env']
    assert 'PATH' in env
    assert 'PYTHONPATH' in env

    path = env['PATH'].split(os.pathsep)
    logger.debug('Destination PATH=%s', path)
    assert TESTSDIR in path
    toxbindir = os.getenv('TOXBINDIR')
    if toxbindir:
        assert toxbindir in path
        assert path.index(toxbindir) < path.index(TESTSDIR)

    pypath = env['PYTHONPATH'].split(os.pathsep)
    logger.debug('Destination PYTHONPATH=%s', path)
    if not toxbindir:
        assert SRCDIR in pypath



from otpush import push, receive

import os
import pytest
import shutil
import subprocess
import yaml

from .util import (
    DATADIR,
    PGP_KEY,
    TESTSDIR,
    SRCDIR,
    kill_gpg_agent,
    ssh_server,
    TmpRepo,
)


@pytest.fixture(scope='session')
def cli_env_vars():
    """Provide environment variables for running via CLI

    Provide adjusted PATH and PYTHONPATH environment variables for the
    processes run subprocess or sshd so that they more closely matches
    the current environment.
    """
    env_vars = {}

    path = os.getenv('PATH', os.defpath)
    cli_path_parts = path.split(os.pathsep)
    cli_path_parts.insert(0, TESTSDIR)
    toxbindir = os.getenv('TOXBINDIR')
    if toxbindir:
        cli_path_parts.insert(0, toxbindir)
    cli_path = os.pathsep.join(cli_path_parts)
    env_vars['PATH'] = cli_path

    pypath = os.getenv('PYTHONPATH')
    cli_pypath_parts = pypath.split(os.pathsep) if pypath else []
    if not toxbindir:
        cli_pypath_parts.insert(0, SRCDIR)
    cli_pypath = os.pathsep.join(cli_pypath_parts)
    env_vars['PYTHONPATH'] = cli_pypath

    return env_vars


@pytest.fixture(scope='session')
def ssh_datadir(tmp_path_factory):
    datadir = tmp_path_factory.mktemp('ssh-data')
    for src in ('host_rsa_key', 'host_rsa_key.pub', 'id_rsa',
                'id_rsa.pub', 'sshd_config'):
        shutil.copy(os.path.join(DATADIR, src), datadir)

    # ssh and sshd refuse to start when the private keys are group or
    # world accessible.
    os.chmod(datadir / 'host_rsa_key', 0o600)
    os.chmod(datadir / 'id_rsa', 0o600)

    return datadir


@pytest.fixture(scope='session')
def sshd(ssh_datadir, cli_env_vars):
    sshd_config = ssh_datadir / 'sshd_config'
    host_key = ssh_datadir / 'host_rsa_key'
    authorized_keys = ssh_datadir / 'id_rsa.pub'
    with ssh_server(sshd_config, host_key, authorized_keys, cli_env_vars) \
         as server_info:
        yield server_info


@pytest.fixture
def ssh_options(sshd, ssh_datadir):
    id_file = ssh_datadir / 'id_rsa'
    return [
        '-i', str(id_file),
        '-o', 'StrictHostKeyChecking=no',
        '-o', 'UserKnownHostsFile=/dev/null',
    ]


@pytest.fixture
def ssh_socket(tmp_path):
    return str(tmp_path / 'ssh_socket')


@pytest.fixture
def tmp_files_path(tmp_path):
    return tmp_path / 'files'


@pytest.fixture
def source_repo(tmp_path, tmp_files_path):
    repo_path = tmp_path / 'source-repo'
    repo = TmpRepo(repo_path, collection_id=None)

    # Turn on auto summary generation
    config = repo.copy_config()
    config.set_boolean('core', 'auto-update-summary', True)
    repo.write_config(config)

    return repo


@pytest.fixture
def source_server(source_repo):
    with push.RepoServer(source_repo.path) as server:
        yield server


@pytest.fixture
def dest_repo(tmp_path):
    repo_path = tmp_path / 'dest-repo'
    return TmpRepo(repo_path)


@pytest.fixture
def receive_repo(dest_repo, source_server):
    repo_path = str(dest_repo.path)
    config = receive.OTReceiveConfig(update=False)
    with receive.OTReceiveRepo(repo_path, source_server.url, config) as repo:
        yield repo


@pytest.fixture
def receive_config_path(tmp_path):
    path = tmp_path / 'ostree-receive.conf'
    config = {
        'update': False,
    }
    with path.open('w') as f:
        yaml.dump(config, f)
    return path


@pytest.fixture
def gpg_homedir(tmp_path):
    """Temporary GPG homedir with private key imported"""
    homedir = tmp_path / 'gnupg'
    homedir.mkdir(mode=0o700)

    cmd = ('gpg', '--homedir', str(homedir), '--import', str(PGP_KEY))
    subprocess.run(cmd, check=True)

    yield homedir

    kill_gpg_agent(homedir)



from otpush import push

import argparse
from gi.repository import GLib, Gio
import logging
import json
import os
from pathlib import Path
import pytest
import re
import socket
import subprocess
from urllib.error import HTTPError
from urllib.request import urlopen

from .util import (
    TESTSDIR,
    needs_sshd,
    random_commit,
    TmpRepo,
)

logger = logging.getLogger(__name__)


class TestRepoServer:
    def populate_repo(self):
        self.repo.mkdir()
        sub = self.repo / 'sub'
        sub.mkdir()
        with open(self.repo / 'a', 'w') as f:
            f.write('foo')
        with open(sub / 'b', 'w') as f:
            f.write('bar')

    def check_server(self):
        assert self.server.path == self.repo
        assert self.server.proc.pid > 0
        assert self.server.address[0] == '127.0.0.1'
        assert self.server.address[1] > 0
        assert self.server.url.startswith('http://127.0.0.1:')
        with urlopen(f'{self.server.url}/a') as resp:
            assert resp.read().decode('utf-8') == 'foo'
        with urlopen(f'{self.server.url}/sub/b') as resp:
            assert resp.read().decode('utf-8') == 'bar'
        with pytest.raises(HTTPError) as excinfo:
            urlopen(f'{self.server.url}/missing')
        assert excinfo.value.getcode() == 404

    def test_missing(self, tmp_path):
        # Nonexistent directory should fail
        repo = tmp_path / 'repo'
        server = None
        with pytest.raises(ValueError) as excinfo:
            server = push.RepoServer(repo)
        assert str(excinfo.value) == f'{repo} is not a directory'
        assert server is None

    def test_non_context(self, tmp_path):
        # Without context manager
        self.repo = tmp_path / 'repo'
        self.populate_repo()

        self.server = push.RepoServer(self.repo)
        assert self.server.path == self.repo
        assert self.server.proc is None
        assert self.server.address is None
        assert self.server.url is None

        # This should do nothing
        self.server.stop()

        # Make sure to clean up so the tests don't hang if there are failures
        try:
            self.server.start()
            self.check_server()
            self.server.stop()
        finally:
            self.server.stop()

    def test_context(self, tmp_path):
        self.repo = tmp_path / 'repo'
        self.populate_repo()
        with push.RepoServer(self.repo) as self.server:
            self.check_server()

    def test_pull(self, tmp_path, tmp_files_path):
        local_repo = TmpRepo(tmp_path / 'local')
        remote_repo = TmpRepo(tmp_path / 'remote')
        random_commit(remote_repo, tmp_files_path, 'test')
        _, remote_refs = remote_repo.list_refs()

        with push.RepoServer(str(remote_repo.path)) as remote_server:
            repo_options = GLib.Variant('a{sv}', {
                'gpg-verify': GLib.Variant('b', False),
                'gpg-verify-summary': GLib.Variant('b', False),
            })
            local_repo.remote_add('origin', remote_server.url, repo_options)

            # Pulling a missing ref should fail
            pull_options = GLib.Variant('a{sv}', {
                'refs': GLib.Variant('as', ['missing']),
            })
            with pytest.raises(GLib.Error) as excinfo:
                local_repo.pull_with_options('origin', pull_options)
            assert excinfo.value.matches(Gio.io_error_quark(),
                                         Gio.IOErrorEnum.NOT_FOUND)
            _, local_refs = local_repo.list_refs()
            assert local_refs == {}

            # Pulling the existing ref should succeed
            pull_options = GLib.Variant('a{sv}', {
                'refs': GLib.Variant('as', ['test']),
            })
            local_repo.pull_with_options('origin', pull_options)
            _, local_refs = local_repo.list_refs()
            assert local_refs == {'origin:test': remote_refs['test']}

            # Pulling with all refs should fail because no branches were
            # setup in the configuration
            local_repo.set_ref_immediate('origin', 'test', None)
            _, local_refs = local_repo.list_refs()
            assert local_refs == {}
            pull_options = GLib.Variant('a{sv}', {})
            with pytest.raises(GLib.Error) as excinfo:
                local_repo.pull_with_options('origin', pull_options)
            assert excinfo.value.matches(Gio.io_error_quark(),
                                         Gio.IOErrorEnum.FAILED)
            assert ('No configured branches for remote origin'
                    in str(excinfo.value))
            assert local_refs == {}


@needs_sshd
class TestSSHMultiplexer:
    def test_socket_exists(self, sshd, ssh_options, ssh_socket):
        ssh = push.SSHMultiplexer(sshd.address, ssh_socket, ssh_options,
                                  port=sshd.port)
        with open(ssh_socket, 'w'):
            pass
        with pytest.raises(push.OTPushError) as excinfo:
            ssh.start()
        assert str(excinfo.value) == f'Socket {ssh_socket} already exists'

    def test_master_non_context(self, sshd, ssh_options, ssh_socket):
        ssh = push.SSHMultiplexer(sshd.address, ssh_socket, ssh_options,
                                  port=sshd.port)

        # Stopping without starting should do nothing
        assert ssh.master_proc is None
        assert not os.path.exists(ssh_socket)
        ssh.stop()

        try:
            ssh.start()
            assert ssh.master_proc.pid > 0
            assert os.path.exists(ssh_socket)

            with pytest.raises(push.OTPushError) as excinfo:
                ssh.start()
            assert str(excinfo.value).startswith(
                'SSH master process already running')
        finally:
            ssh.stop()

    def test_master_context(self, sshd, ssh_options, ssh_socket):
        with push.SSHMultiplexer(sshd.address, ssh_socket, ssh_options,
                                 port=sshd.port) as ssh:
            assert ssh.master_proc.pid > 0
            assert os.path.exists(ssh_socket)

    def test_forward_port(self, sshd, ssh_options, ssh_socket):
        sock1 = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        sock1.bind(('127.0.0.1', 0))
        sock1_port = sock1.getsockname()[1]
        assert sock1_port > 0

        ssh = push.SSHMultiplexer(sshd.address, ssh_socket, ssh_options,
                                  port=sshd.port)
        with pytest.raises(push.OTPushError) as excinfo:
            ssh.forward_port(sock1_port)
        assert str(excinfo.value) == 'SSH master process not running'
        assert ssh.master_proc is None

        with push.SSHMultiplexer(sshd.address, ssh_socket, ssh_options,
                                 port=sshd.port) as ssh:
            sock2_port = ssh.forward_port(sock1_port)
            assert sock2_port > 0

            sock3 = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            sock3.connect(('127.0.0.1', sock2_port))

    def test_run(self, tmp_path, sshd, ssh_options, ssh_socket):
        ssh = push.SSHMultiplexer(sshd.address, ssh_socket, ssh_options,
                                  port=sshd.port)
        with pytest.raises(push.OTPushError) as excinfo:
            ssh.run(['true'])
        assert str(excinfo.value) == 'SSH master process not running'
        assert ssh.master_proc is None

        with push.SSHMultiplexer(sshd.address, ssh_socket, ssh_options,
                                 port=sshd.port) as ssh:
            test_file = tmp_path / 'test_file'
            assert not os.path.exists(test_file)
            ssh.run(['touch', str(test_file)])
            assert os.path.exists(test_file)

            with pytest.raises(subprocess.CalledProcessError,
                               match='returned non-zero exit status 1'):
                ssh.run(['rmdir', str(tmp_path)])


@needs_sshd
class TestPushRefs:
    DUMPENV_PATH = os.path.join(TESTSDIR, 'dumpenv')

    def push_refs(self, source_repo, dest_repo, sshd, ssh_options, capfd,
                  refs=None, dry_run=False):
        """Run push.push_refs and check the remote command is correct"""
        dest = push.PushDest(host=sshd.address, port=sshd.port,
                             repo=str(dest_repo.path), user=None)
        push.push_refs(source_repo, dest, refs=refs, dry_run=dry_run,
                       ssh_options=ssh_options, command='dumpenv')

        out, _ = capfd.readouterr()
        data = json.loads(out)
        args = data['args']

        num_args = len(args)
        num_refs = len(refs) if refs else 0
        expected_num_args = num_refs + 3
        if dry_run:
            expected_num_args += 1
        assert num_args == expected_num_args

        args_iter = iter(args)
        assert next(args_iter) == self.DUMPENV_PATH
        if dry_run:
            assert next(args_iter) == '-n'
        assert next(args_iter) == str(dest_repo.path)
        assert next(args_iter).startswith('http://127.0.0.1:')
        remaining = list(args_iter)
        if refs:
            assert remaining == refs
        else:
            assert remaining == []

    def test_no_refs(self, source_repo, dest_repo, sshd, ssh_options,
                     tmp_files_path, capfd):
        self.push_refs(source_repo, dest_repo, sshd, ssh_options, capfd,
                       refs=None)

        self.push_refs(source_repo, dest_repo, sshd, ssh_options, capfd,
                       refs=[])

    def test_refs(self, source_repo, dest_repo, sshd, ssh_options,
                  tmp_files_path, capfd):
        random_commit(source_repo, tmp_files_path, 'test1')
        random_commit(source_repo, tmp_files_path, 'test2')
        self.push_refs(source_repo, dest_repo, sshd, ssh_options, capfd,
                       refs=['test1'])
        self.push_refs(source_repo, dest_repo, sshd, ssh_options, capfd,
                       refs=['test2'])
        self.push_refs(source_repo, dest_repo, sshd, ssh_options, capfd,
                       refs=['test1', 'test2'])

    def test_missing_ref(self, source_repo, dest_repo, sshd, ssh_options,
                         tmp_files_path, capfd):
        random_commit(source_repo, tmp_files_path, 'test')
        with pytest.raises(push.OTPushError) as excinfo:
            self.push_refs(source_repo, dest_repo, sshd, ssh_options, capfd,
                           refs=['missing'])
        assert str(excinfo.value) == \
            f'Refs missing not found in {source_repo.path}'
        with pytest.raises(push.OTPushError) as excinfo:
            self.push_refs(source_repo, dest_repo, sshd, ssh_options, capfd,
                           refs=['test', 'missing'])
        assert str(excinfo.value) == \
            f'Refs missing not found in {source_repo.path}'

    def test_summary(self, source_repo, dest_repo, sshd, ssh_options,
                     tmp_files_path, capfd):
        summary = Path(source_repo.path) / 'summary'
        random_commit(source_repo, tmp_files_path, 'test')

        # Delete the summary file and check that it gets generated.
        summary.unlink()
        self.push_refs(source_repo, dest_repo, sshd, ssh_options, capfd)
        assert summary.exists()

        # Set the summary mtime behind the repo and check that it gets
        # regenerated.
        repo_mtime = os.path.getmtime(source_repo.path)
        os.utime(summary, (repo_mtime - 1, repo_mtime - 1))
        orig_summary_mtime = summary.stat().st_mtime
        assert orig_summary_mtime < repo_mtime
        self.push_refs(source_repo, dest_repo, sshd, ssh_options, capfd)
        assert summary.exists()
        new_summary_mtime = summary.stat().st_mtime
        assert new_summary_mtime > orig_summary_mtime

    def test_dry_run(self, source_repo, dest_repo, sshd, ssh_options,
                     tmp_files_path, capfd):
        self.push_refs(source_repo, dest_repo, sshd, ssh_options, capfd,
                       dry_run=True)


class TestParseDest:
    def test_bad_scheme(self):
        for scheme in ('http', 'ftp', 'scp', 'blah'):
            with pytest.raises(
                    ValueError,
                    match=f'Destination scheme "{scheme}" not allowed'):
                push.parse_dest(f'{scheme}://host/repo')

    def test_missing_repo(self):
        for dest in ('ssh://', 'http://', 'host:', 'user@host:'):
            with pytest.raises(ValueError, match='Destination repo missing'):
                push.parse_dest('host:')

    def test_empty_dest(self):
        with pytest.raises(ValueError,
                           match='Destination not in form "host:repo"'):
            push.parse_dest('')

    def test_missing_host(self):
        for dest in (':', ':repo', ':/path/:/repo'):
            with pytest.raises(ValueError, match='Destination host missing'):
                push.parse_dest(dest)

    def test_invalid_host(self):
        for dest in ('@:repo', '@host:repo', 'user@:repo'):
            with pytest.raises(ValueError, match='Invalid destination host'):
                push.parse_dest(dest)

    def test_invalid_port(self):
        match = re.compile(
            r'(Port could not be cast to integer|invalid literal for int)'
        )
        for dest in ('ssh://host:port/repo', 'ssh://host:$/repo'):
            with pytest.raises(ValueError, match=match):
                push.parse_dest(dest)

    def test_good_dest(self):
        cases = (
            ('ssh://host/repo',
             push.PushDest(host='host', repo='/repo', user=None, port=None)),
            ('ssh://host.example.com/repo',
             push.PushDest(host='host.example.com', repo='/repo', user=None,
                           port=None)),
            ('ssh://host/path/to/repo/',
             push.PushDest(host='host', repo='/path/to/repo/', user=None,
                           port=None)),
            ('ssh://host/path/:/repo',
             push.PushDest(host='host', repo='/path/:/repo', user=None,
                           port=None)),
            ('ssh://user@host/repo',
             push.PushDest(host='host', user='user', repo='/repo', port=None)),
            ('ssh://host:22/repo',
             push.PushDest(host='host', port=22, repo='/repo', user=None)),
            ('host:repo',
             push.PushDest(host='host', repo='repo', user=None, port=None)),
            ('host:path/to/repo',
             push.PushDest(host='host', repo='path/to/repo', user=None,
                           port=None)),
            ('host:/repo',
             push.PushDest(host='host', repo='/repo', user=None, port=None)),
            ('host:/path/:/repo',
             push.PushDest(host='host', repo='/path/:/repo', user=None,
                           port=None)),
            ('user@host:repo',
             push.PushDest(host='host', user='user', repo='repo', port=None)),
            ('user@host.example.com:path/to/repo',
             push.PushDest(host='host.example.com', user='user',
                           repo='path/to/repo', port=None)),
        )

        for arg, expected in cases:
            dest = push.parse_dest(arg)
            assert dest == expected


class TestArgParser:
    def test_no_dest(self, capsys):
        ap = push.OTPushArgParser()
        with pytest.raises(SystemExit) as excinfo:
            ap.parse_args([])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert out == ''
        assert err.endswith('error: the following arguments are required: '
                            'DEST\n')

    def test_defaults(self):
        ap = push.OTPushArgParser()
        args = ap.parse_args(['host:repo'])
        assert args == argparse.Namespace(
            command='ostree-receive',
            dest=push.PushDest(host='host', repo='repo', user=None, port=None),
            dry_run=False,
            log_level=logging.INFO,
            port=None,
            refs=[],
            repo=None,
            ssh_options=[],
        )

    def test_dest(self):
        ap = push.OTPushArgParser()
        args = ap.parse_args(['host:repo'])
        assert args.dest == push.PushDest(host='host', repo='repo',
                                          user=None, port=None)
        args = ap.parse_args(['user@host:repo'])
        assert args.dest == push.PushDest(host='host', user='user',
                                          repo='repo', port=None)
        args = ap.parse_args(['ssh://user@host/repo'])
        assert args.dest == push.PushDest(host='host', user='user',
                                          repo='/repo', port=None)
        args = ap.parse_args(['ssh://user@host:1234/repo'])
        assert args.dest == push.PushDest(host='host', user='user',
                                          port=1234, repo='/repo')

    def test_refs(self):
        ap = push.OTPushArgParser()
        args = ap.parse_args(['host:repo', 'foo'])
        assert args.refs == ['foo']
        args = ap.parse_args(['host:repo', 'foo', 'bar', 'baz'])
        assert args.refs == ['foo', 'bar', 'baz']

    def test_port(self, capsys):
        ap = push.OTPushArgParser()
        args = ap.parse_args(['-p', '22', 'host:repo'])
        assert args.port == 22

        with pytest.raises(SystemExit) as excinfo:
            ap.parse_args(['-p', 'foo', 'host:repo'])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert out == ''
        assert err.endswith("invalid int value: 'foo'\n")

    def test_port_and_dest_port(self):
        ap = push.OTPushArgParser()
        args = ap.parse_args(['-p', '22', 'ssh://host:2200/repo'])
        assert args.port == 22
        assert args.dest.port == 22

    def test_dry_run(self):
        ap = push.OTPushArgParser()
        args = ap.parse_args(['-n', 'host:repo'])
        assert args.dry_run is True
        args = ap.parse_args(['--dry-run', 'host:repo'])
        assert args.dry_run is True

    def test_log_level(self):
        ap = push.OTPushArgParser()
        args = ap.parse_args(['-v', 'host:repo'])
        assert args.log_level == logging.DEBUG
        args = ap.parse_args(['--verbose', 'host:repo'])
        assert args.log_level == logging.DEBUG
        args = ap.parse_args(['-q', 'host:repo'])
        assert args.log_level == logging.WARNING
        args = ap.parse_args(['--quiet', 'host:repo'])
        assert args.log_level == logging.WARNING

    def test_repo(self):
        ap = push.OTPushArgParser()
        args = ap.parse_args(['--repo', '/repo', 'host:repo'])
        assert args.repo == '/repo'

    def test_command(self):
        ap = push.OTPushArgParser()
        args = ap.parse_args(['--command=ls', 'host:repo'])
        assert args.command == 'ls'
        args = ap.parse_args(['--command=ostree-receive', 'host:repo'])
        assert args.command == 'ostree-receive'
        args = ap.parse_args(['--command', '/path/to/ostree-receive',
                              'host:repo'])
        assert args.command == '/path/to/ostree-receive'

    def test_ssh_options(self, capsys):
        ap = push.OTPushArgParser()
        args = ap.parse_args(['-ifoo', 'host:repo'])
        assert args.ssh_options == ['-i', 'foo']
        args = ap.parse_args(['-i', 'foo', 'host:repo'])
        assert args.ssh_options == ['-i', 'foo']
        args = ap.parse_args(['-o', 'Foo=yes', 'host:repo'])
        assert args.ssh_options == ['-o', 'Foo=yes']
        args = ap.parse_args(['-o', 'Foo=yes', '-o', 'Bar=no',
                              'host:repo'])
        assert args.ssh_options == ['-o', 'Foo=yes', '-o', 'Bar=no']

        with pytest.raises(SystemExit) as excinfo:
            ap.parse_args(['-ifoo', '-ibar', 'host:repo'])
        assert excinfo.value.code == 2
        out, err = capsys.readouterr()
        assert out == ''
        assert err.endswith('Option -i can only be specified once\n')



from collections import namedtuple
from contextlib import contextmanager
import getpass
import gi
import logging
import os
import pytest
import random
import shutil
import socket
import subprocess
import time

gi.require_version('OSTree', '1.0')
from gi.repository import GLib, Gio, OSTree  # noqa: E402

logger = logging.getLogger(__name__)
TESTSDIR = os.path.abspath(os.path.dirname(__file__))
DATADIR = os.path.join(TESTSDIR, 'data')
SRCDIR = os.path.dirname(TESTSDIR)
SCRIPTSDIR = os.path.join(SRCDIR, 'scripts')
PGP_KEY = os.path.join(DATADIR, 'pgp-key.asc')
PGP_PUB = os.path.join(DATADIR, 'pgp-pub.asc')
PGP_PUB_KEYRING = os.path.join(DATADIR, 'pgp-pub.gpg')
PGP_KEY_ID = '281D0DDC6EDD77CF6A8A936C247D3E51CDA08B6B'


class OTPushTestError(Exception):
    pass


@contextmanager
def oneshot_transaction(repo):
    """Start an OSTree repo transaction and abort on any failures"""
    repo.prepare_transaction()
    try:
        yield
        # Commit the transaction
        repo.commit_transaction()
    except:  # noqa: E722
        # Abort on any failures
        repo.abort_transaction()
        raise


needs_ostree = pytest.mark.skipif(
    not shutil.which('ostree'), reason='ostree required'
)


needs_flatpak = pytest.mark.skipif(
    not shutil.which('flatpak'), reason='flatpak required'
)


def random_commit(repo, tmpdir, refspec, parent=None, timestamp=None,
                  extra_metadata=None):
    """Create a random commit and set refspec to it

    Returns the new commit checksum.
    """
    for path in ('a', 'sub/b'):
        full_path = tmpdir / path
        full_path.parent.mkdir(exist_ok=True)
        rand_size = random.randrange(1000, 10000)
        with open(full_path, 'wb') as f:
            f.write(os.urandom(rand_size))

    # Use current UTC time if no timestamp specified
    if timestamp is None:
        timestamp = int(time.time())
    logger.info('Using timestamp %u for random commit on %s', timestamp,
                refspec)

    # Include the collection and ref bindings in the commit
    metadata = {}
    if extra_metadata:
        metadata.update(extra_metadata)
    collection_id = repo.get_collection_id()
    if collection_id is not None:
        metadata[OSTree.COMMIT_META_KEY_COLLECTION_BINDING] = \
                GLib.Variant('s', collection_id)
    _, remote, ref = OSTree.parse_refspec(refspec)
    metadata[OSTree.COMMIT_META_KEY_REF_BINDING] = GLib.Variant('as', [ref])
    metadata_var = GLib.Variant('a{sv}', metadata)

    with oneshot_transaction(repo):
        # Populate a mutable tree from the random files
        mtree = OSTree.MutableTree.new()
        repo.write_directory_to_mtree(Gio.File.new_for_path(str(tmpdir)),
                                      mtree, None)
        _, root = repo.write_mtree(mtree)

        # Commit the mtree root
        _, checksum = repo.write_commit_with_time(parent,
                                                  'Test commit',
                                                  None,
                                                  metadata_var,
                                                  root,
                                                  timestamp)

        # Set the ref
        if remote is None:
            # Local ref, set a collection ref
            collection_ref = OSTree.CollectionRef()
            collection_ref.collection_id = collection_id
            collection_ref.ref_name = ref
            repo.transaction_set_collection_ref(collection_ref, checksum)
        else:
            # Remote ref
            repo.transaction_set_ref(remote, ref, checksum)

    logger.info('Created random commit %s on %s', checksum, refspec)
    return checksum


def wipe_repo(repo):
    """Delete all refs and objects in repo"""
    _, refs = repo.list_refs(None)
    for refspec in refs.keys():
        _, remote, ref = OSTree.parse_refspec(refspec)
        repo.set_ref_immediate(remote, ref, None)
    repo.prune(OSTree.RepoPruneFlags.REFS_ONLY, -1)


def local_refs(repo, prefix=None):
    """Get local refs in repo excluding remotes and mirrors"""
    flags = OSTree.RepoListRefsExtFlags.EXCLUDE_REMOTES
    try:
        # EXCLUDE_MIRRORS only available since ostree 2019.2
        flags |= OSTree.RepoListRefsExtFlags.EXCLUDE_MIRRORS
    except AttributeError:
        pass
    _, refs = repo.list_refs_ext(prefix, flags)
    return refs


def get_summary_variant(path):
    summary_file = Gio.File.new_for_path(str(path))
    summary_bytes, _ = summary_file.load_bytes()
    summary_variant = GLib.Variant.new_from_bytes(
        type=GLib.VariantType(OSTree.SUMMARY_GVARIANT_STRING),
        bytes=summary_bytes, trusted=False)
    if not summary_variant.is_normal_form():
        raise OTPushTestError(
            f'Summary file {path} is not in normal GVariant format')
    return summary_variant


def get_content_checksum(repo, rev):
    """Get a commit's content checksum"""
    _, commit, _ = repo.load_commit(rev)
    return OSTree.commit_get_content_checksum(commit)


class TmpRepo(OSTree.Repo):
    """Temporary OSTree repo"""
    COLLECTION_ID = 'com.example.Test'

    def __init__(self, path, collection_id=COLLECTION_ID, **kwargs):
        self.path = path

        self.path.mkdir()
        repo_file = Gio.File.new_for_path(str(self.path))
        super().__init__(path=repo_file, **kwargs)
        if collection_id:
            self.set_collection_id(collection_id)
        self.create(OSTree.RepoMode.ARCHIVE)


SSHServerInfo = namedtuple('SSHServerInfo', ('proc', 'address', 'port'))


def get_sshd():
    """Returns the path to sshd or None

    Looks in PATH and typical sbin directories not in PATH.
    """
    path = os.getenv('PATH', os.defpath)
    sshd_path = os.pathsep.join([path, '/usr/local/sbin', '/usr/sbin',
                                 '/sbin'])
    sshd = shutil.which('sshd', path=sshd_path)
    if sshd:
        logger.debug('Found sshd %s', sshd)
    else:
        logger.debug('sshd not found in %s', sshd_path)
    return sshd


def have_required_sshd(sshd=None):
    """Check whether sshd meets requirements

    sshd needs to be OpenSSH version 7.8 or newer to support the SetEnv
    config option.
    """
    if not sshd:
        sshd = get_sshd()
    if not sshd:
        return False

    # Run sshd -V to get the version. This is actually only an option on
    # the ssh client, but it will print the version after complaining
    # about the unknown option. Maybe someday it will exist...
    proc = subprocess.run([sshd, '-V'], stdout=subprocess.PIPE,
                          stderr=subprocess.STDOUT)
    lines = iter(proc.stdout.decode('utf-8').splitlines())
    try:
        version_line = next(lines)
        if version_line.startswith('unknown option'):
            version_line = next(lines)
    except StopIteration:
        logger.debug('No version information from %s -V', sshd)
        return False

    logger.debug('sshd version line: %s', version_line)
    if not version_line.startswith('OpenSSH_'):
        logger.debug('OpenSSH not in version')
        return False

    # The version line should look something like:
    #
    # OpenSSH_7.6p1 Ubuntu-4ubuntu0.3, OpenSSL 1.0.2n  7 Dec 2017
    #
    # Get the first word, strip the OpenSSH_ prefix, strip the portable
    # pX suffix, and then try to get the major and minor version
    # numbers.
    openssh_version = version_line.split()[0]
    _, _, full_version = openssh_version.partition('OpenSSH_')
    version, _, _ = full_version.partition('p')
    version_parts = version.split('.')
    try:
        major = int(version_parts[0])
    except ValueError:
        logger.debug('Could not get major version from %s', version)
        return False
    try:
        minor = int(version_parts[1])
    except IndexError:
        minor = 0
    except ValueError:
        logger.debug('Could not get minor version from %s', version)
        return False
    logger.debug('Detected OpenSSH sshd version %d.%d', major, minor)

    # OpenSSH 7.8 is needed for the SetEnv option
    if major < 7 or (major == 7 and minor < 8):
        logger.debug('OpenSSH sshd version < 7.8')
        return False

    return True


needs_sshd = pytest.mark.skipif(
    not have_required_sshd(),
    reason='OpenSSH sshd version 7.8+ required'
)


def get_ssh_server_id(address):
    """Open a connection to an SSH server and get the identification string

    If a connection could not be established or no data is received, an
    empty string is returned.
    """
    for _ in range(10):
        try:
            sock = socket.create_connection(address, timeout=1)
            break
        except ConnectionRefusedError:
            logger.debug('Could not connect to port %d', address[1])
            time.sleep(0.01)
    else:
        logger.debug('Could not connect in 5 attempts')
        return ''

    try:
        return sock.recv(256).decode('utf-8')
    except socket.timeout:
        logger.debug('No data received from port %d', address[1])
        return ''
    finally:
        sock.close()


@contextmanager
def ssh_server(sshd_config, host_key, authorized_keys, env_vars=None):
    # Running sshd requires an absolute path
    sshd = get_sshd()
    if not sshd:
        raise OTPushTestError('Could not find sshd')
    if not have_required_sshd(sshd):
        raise OTPushTestError(f'{sshd} is not the required version')

    # Build a SetEnv option value from the provided environment variables.
    env_vars = env_vars or {}
    setenv = ' '.join([
        f'{var}="{value}"'
        for var, value in env_vars.items()
    ])

    cmd = [
        sshd,
        # Don't fork
        '-D',
        # Write logs to stderr
        '-e',
        # Config file
        '-f', str(sshd_config),
        # Host key file
        '-h', str(host_key),
        # Authorized keys file
        '-o', f'AuthorizedKeysFile={authorized_keys}',
        # Only allow running user
        '-o', f'AllowUsers={getpass.getuser()}',
        # Allow root login
        '-o', 'PermitRootLogin=yes',
        # Set environment variables for the process
        '-o', f'SetEnv={setenv}',
    ]
    logger.debug('SSH server args: %s', ' '.join(cmd))

    # Loop a few times trying to find an open ephemeral port
    with open('/proc/sys/net/ipv4/ip_local_port_range') as f:
        start_port, stop_port = map(int, f.readline().split())
    proc = None
    try:
        for _ in range(5):
            port = random.randrange(start_port, stop_port)
            logger.debug('Starting %s with port %d', sshd, port)
            proc = subprocess.Popen(cmd + ['-p', str(port)])
            server_id = get_ssh_server_id(('127.0.0.1', port))
            if server_id.startswith('SSH-2.0-'):
                logger.info('%s started on port %d', sshd, port)
                break
        else:
            raise OTPushTestError(f'Could not start {sshd}')

        yield SSHServerInfo(proc, '127.0.0.1', port)
    finally:
        if proc is not None and proc.poll() is None:
            logger.debug('Stopping sshd process %d', proc.pid)
            proc.terminate()


needs_gpg = pytest.mark.skipif(
    not shutil.which('gpg'), reason='gpg required'
)


def get_gnupg_version():
    """Get the GnuPG version from gpg"""
    proc = subprocess.run(['gpg', '--version'], stdout=subprocess.PIPE,
                          check=True)
    lines = proc.stdout.decode('utf-8').splitlines()
    if len(lines) == 0:
        raise OTPushTestError(
            'No version information from gpg --version'
        )

    version = lines[0].split()[-1]
    logger.debug('Found GnuPG version %s', version)
    return version


def kill_gpg_agent(gpg_homedir):
    """Kill gpg-agent in GPG homedir"""
    version = get_gnupg_version()
    version_parts = version.split('.')
    if len(version_parts) < 3:
        raise OTPushTestError(
            f'GnuPG version {version} has less than 3 components'
        )
    major, minor, patch = map(int, version_parts[0:3])

    # If gpg is at least version 2.1.17, gpg-agent (properly) watches
    # the homedir with an inotify watch and exits when it's deleted.
    if (
        major > 2 or
        (major == 2 and minor > 1) or
        (major == 2 and minor == 1 and patch > 17)
    ):
        logger.debug('GnuPG >= 2.1.17, skipping gpg-agent cleanup')
        return

    # Otherwise, it needs to be told to exit. Note that this is
    # asynchronous. If the caller is deleting the homedir, there can be
    # failures as both the caller and gpg-agent try to delete the
    # sockets concurrently.
    cmd = (
        'gpg-connect-agent',
        '--homedir',
        str(gpg_homedir),
        'killagent',
        '/bye'
    )
    logger.debug('Killing GnuPG agent: %s', ' '.join(cmd))
    subprocess.run(cmd, check=True)
